# coding:utf-8
require 'rubygems'
require 'socket'
require 'thread'
require 'pp'
require 'yaml'
require 'logger'
require 'fileutils'
require 'securerandom'
require 'net/https'
require_relative 'lib/claim_xml'

class ClaimReceiver

VERSION = "0.0.5"

EOT = 0x04.chr
ACK = 0x06.chr
NAK = 0x15.chr
ERR = 0x20.chr

H_CODE = {
EOT => "EOT",
ACK => "ACK",
NAK => "NAK",
ERR => "ERR"
}

  def initialize(conf)
    @conf = YAML.load(File.read(conf))
    @logger = Logger.new(@conf[:log_file],7,10*1024*1024)
	case @conf[:log_level]
	when "debug"
	  @logger.level = Logger::DEBUG
	else
	  @logger.level = Logger::INFO
	end 
  end

  def start
    @logger.info "start"
    port = @conf[:port] || 11111
    server = TCPServer.new port
    loop do
      Thread.start(server.accept) do |s|
        begin
          @logger.info "connect #{s.peeraddr}"
          xml = ""
     	  loop do
            c = s.recv(1)
            if c == EOT
              break
            else
              xml += c
            end
          end
          @logger.debug xml

          # write xml
          dir = @conf[:claim_xml_dir]
          if @conf[:claim_xml_dir]
            limit = @conf[:claim_xml_limit] || 100
            fname = Time.now.strftime("%Y%m%d%H%M%S-#{SecureRandom.uuid}.xml")
            fname = File.join(dir,fname)
            file_logger(dir,fname,limit) do
              File.open(fname,"wb") do |io| io.print xml end
            end
            @logger.info "write #{fname}"
          end
          res = send_claim(xml)
          @logger.info "res: #{H_CODE[res]}"
          s.putc res
          s.close
        rescue Exception => ex
          @logger.error(ex)
          @logger.error(ex.backtrace)
        end
      end
    end
  end

  private

  def send_claim(xml)
    begin
      data = make_request_data(xml)
      unless data
        @logger.info "invalid xml"
        return NAK
      end
      unless data["claim_receivev2req"]["Patient_ID"]
        @logger.info "invalid xml. Patient_ID is nil"
        return NAK
      end
      json = data.to_json
      @logger.debug json
      res = http_post(@conf[:api_path],json)
      unless res.code == "200"
        raise "http error #{res.code}"
      end
      @logger.info res.body
      json = JSON.parse(res.body)
      claim_res = json["claim_receivev2res"]
      raise "invalid response json" unless claim_res
      if claim_res["Api_Result"] == "00"
        return ACK
      else
        @logger.info "Api_Result Error"
        @logger.info res
        return NAK
      end
    rescue Exception => ex
      @logger.info "#{@conf[:api_path]} error"
      @logger.info ex
      @logger.info ex.backtrace
      return NAK  
    end
  end

  def http_post(path,data=nil)
    Net::HTTP.version_1_2
    https = Net::HTTP.new(@conf[:api_server],@conf[:api_port])
    if @conf[:use_ssl]
      https.use_ssl = true
      https.ca_file = @conf[:ca_cert]
      https.cert = OpenSSL::X509::Certificate.new(File.read(@conf[:cert]))
      https.key = OpenSSL::PKey::RSA.new(File.read(@conf[:cert_key]),@conf[:passphrase])
      https.verify_mode = OpenSSL::SSL::VERIFY_PEER
      https.verify_depth = 5
    end
  
    https.start do
      req = nil
      if data
        req = Net::HTTP::Post.new(path)
        req.content_length = data.size
        req.content_type = "application/json"
        req.body = data
      else
        req = Net::HTTP::Get.new(path)
      end
      req.basic_auth(@conf[:api_user],@conf[:api_key])
      res = https.request(req)
    end
  end
  
  def file_logger(dir,fname,limit)
     FileUtils.mkdir_p(dir)
     yield
     if limit > 0
      list = Dir.glob(File.join(dir,"*")).sort do |a,b|
        File.ctime(a) <=> File.ctime(b)
      end
      if list.size > limit
        FileUtils.rm_f(list[0,list.size - limit])
      end
    end
  end
  
end

if $0 == __FILE__
  Signal.trap(:INT) do
    exit 1
  end
  
  claim = ClaimReceiver.new(ARGV[0])
  claim.start
end
